from typing import Self, SupportsIndex
from .vector2 import Vector2

class MVector2:
    @staticmethod
    def zero() -> MVector2: ...
    @staticmethod
    def one() -> MVector2: ...
    @staticmethod
    def up() -> MVector2: ...
    @staticmethod
    def down() -> MVector2: ...
    @staticmethod
    def left() -> MVector2: ...
    @staticmethod
    def right() -> MVector2: ...

    x: float
    y: float

    def __init__(self, x: float, y: float) -> None: ...

    @staticmethod
    def from_immutable(immutable: Vector2) -> MVector2: ...

    def __getitem__(self, i: SupportsIndex) -> float: ...

    def __repr__(self) -> str: ...

    def __len__(self) -> int: ...

    def __iadd__(self, other: MVector2 | Vector2) -> Self: ...

    def __isub__(self, other: MVector2 | Vector2) -> Self: ...

    def __imul__(self, other: MVector2 | Vector2 | float | int) -> Self: ...

    def __itruediv__(self, other: MVector2 | Vector2 | float | int) -> Self: ...

    def __ifloordiv__(self, other: MVector2 | Vector2 | float | int) -> Self: ...

    def __imod__(self, other: MVector2 | Vector2 | float | int) -> Self: ...

    def __eq__(self, other: MVector2) -> bool: ...

    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        ...

    def copy(self) -> MVector2:
        """Make a new vector where the components' values match this vectors values."""
        ...

    @property
    def normalized(self) -> MVector2:
        """A copy of this vector with a magnitude of 1"""
        ...

    def normalize(self) -> None:
        """Make this vector keep its direction, but have a magnitude of 1. Operates in place."""
        ...

    @staticmethod
    def min(a: MVector2, b: MVector2) -> MVector2:
        """Create a new `MVector2` by selecting the smallest components of the two given `MVector2` instances."""
        ...

    @staticmethod
    def max(a: MVector2, b: MVector2) -> MVector2:
        """Create a new `MVector2` by selecting the largest components of the two given `MVector2` instances."""
        ...

    def to_immutable(self) -> Vector2: ...
    def to_tuple(self) -> tuple[float, float]: ...
    def to_dict(self) -> dict[str, float]: ...



import unittest

from nalpy.math import Vector2Int, Vector2


class BasicFunctionality(unittest.TestCase):
    def test_index(self):
        v = Vector2Int(3, 7)
        self.assertEqual(v[0], 3)
        self.assertEqual(v[1], 7)
        self.assertRaises(IndexError, lambda: v[3])

        class MockIndexObj:
            def __init__(self, i: int) -> None:
                self.__internal_index = i
            def __index__(self) -> int:
                return self.__internal_index
        self.assertEqual(v[MockIndexObj(0)], 3)
        self.assertEqual(v[MockIndexObj(1)], 7)

    def test_xy(self):
        v = Vector2Int(3, 7)
        self.assertEqual(v.x, 3)
        self.assertEqual(v.y, 7)
        self.assertRaises(AttributeError, lambda: v.z) # type: ignore

    def test_equals(self):
        v = Vector2Int(3, 3)
        self.assertEqual(v, Vector2Int(3, 3))
        self.assertNotEqual(v, Vector2Int(2, 3))
        self.assertNotEqual(v, Vector2Int(3, 2))
        self.assertNotEqual(v, Vector2Int(2, 2))

        self.assertRaises(TypeError, lambda: v == (3, 3))
        self.assertNotEqual(v, None)
        self.assertNotEqual(None, v)

    def test_hash(self):
        values: tuple[tuple[int, int], ...] = (
            (69, 420),
            (420, 69),
            (69, 69),
            (420, 420),

            (0, 0),
            (-1, -1),
            (-69, -69),
            (1000, 1000),
            (-10000, -10000)
        )

        # Vector2Int hash should match tuple hash
        for v in values:
            self.assertEqual(hash(v), hash(Vector2Int(*v)))

    def test_constants(self):
        self.assertEqual(Vector2Int.zero,  Vector2Int(0, 0))
        self.assertEqual(Vector2Int.one,   Vector2Int(1, 1))
        self.assertEqual(Vector2Int.up,    Vector2Int(0, 1))
        self.assertEqual(Vector2Int.down,  Vector2Int(0, -1))
        self.assertEqual(Vector2Int.left,  Vector2Int(-1, 0))
        self.assertEqual(Vector2Int.right, Vector2Int(1, 0))

    def test_repr(self):
        self.assertEqual(repr(Vector2Int(2, 2)), f"Vector2Int(2, 2)")
        self.assertEqual(repr(Vector2Int(69, 420)), f"Vector2Int(69, 420)")

        self.assertEqual(repr(Vector2Int(2, 2)), str(Vector2Int(2, 2)))
        self.assertEqual(repr(Vector2Int(69, 420)), str(Vector2Int(69, 420)))

    def test_addsub(self):
        b: Vector2Int = Vector2Int.one
        self.assertEqual(b + Vector2Int(2, 2), Vector2Int(3, 3))
        self.assertEqual(b + Vector2Int(0, 2), Vector2Int(1, 3))

        self.assertEqual(b - Vector2Int(2, 2), Vector2Int(-1, -1))
        self.assertEqual(b - Vector2Int(0, 2), Vector2Int(1, -1))

    def test_mult(self):
        b: Vector2Int = Vector2Int.one
        self.assertEqual(b * 2, Vector2Int(2, 2))
        self.assertEqual(2 * b, Vector2Int(2, 2))

        self.assertEqual(Vector2Int(3, 4) * b, Vector2Int(3, 4))
        self.assertEqual(b * Vector2Int(3, 4), Vector2Int(3, 4))

    def test_divmod(self):
        self.assertEqual(Vector2Int(9, 9) / 3, Vector2(3, 3))
        self.assertEqual(Vector2Int(9, 9) / Vector2Int(3, 6), Vector2(3.0, 1.5))
        self.assertEqual(Vector2Int(-9, -9) / 3, Vector2(-3, -3))
        self.assertEqual(Vector2Int(-9, -9) / Vector2Int(3, 6), Vector2(-3.0, -1.5))

        self.assertEqual(Vector2Int(10, 10) // 3, Vector2Int(3, 3))
        self.assertEqual(Vector2Int(10, 10) // Vector2Int(3, 6), Vector2Int(3, 1))
        self.assertEqual(Vector2Int(-10, -10) // 3, Vector2Int(-4, -4))
        self.assertEqual(Vector2Int(-10, -10) // Vector2Int(3, 6), Vector2Int(-4, -2))

        self.assertEqual(Vector2Int(10, 10) % 3, Vector2Int(1, 1))
        self.assertEqual(Vector2Int(10, 10) % Vector2Int(3, 5), Vector2Int(1, 0))
        self.assertEqual(Vector2Int(-10, -10) % 3, Vector2Int(2, 2))
        self.assertEqual(Vector2Int(-10, -10) % Vector2Int(3, 5), Vector2Int(2, 0))

        self.assertEqual(divmod(Vector2Int(10, 10), 3), (Vector2Int(3, 3), Vector2Int(1, 1)))
        self.assertEqual(divmod(Vector2Int(10, 10), Vector2Int(3, 5)), (Vector2Int(3, 2), Vector2Int(1, 0)))
        self.assertEqual(divmod(Vector2Int(-10, -10), 3), (Vector2Int(-4, -4), Vector2Int(2, 2)))
        self.assertEqual(divmod(Vector2Int(-10, -10), Vector2Int(3, 5)), (Vector2Int(-4, -2), Vector2Int(2, 0)))

    def test_rest_operators(self):
        self.assertEqual(-Vector2Int.one, Vector2Int(-1, -1))
        self.assertEqual(-Vector2Int.zero, Vector2Int.zero)
        self.assertEqual(-Vector2Int(-3, -3), Vector2Int(3, 3))

        self.assertEqual(abs(Vector2Int(-2, -2)), Vector2Int(2, 2))
        self.assertEqual(abs(Vector2Int(2, -2)), Vector2Int(2, 2))
        self.assertEqual(abs(Vector2Int(-2, 2)), Vector2Int(2, 2))
        self.assertEqual(abs(Vector2Int(2, 2)), Vector2Int(2, 2))

    def test_properties(self):
        self.assertEqual(Vector2Int(3, 6).magnitude, 6.70820393249936908923)
        self.assertEqual(Vector2Int(7, 12).magnitude, 13.89244398944980450843)

    def test_iter(self):
        i1 = iter(Vector2Int(69, 420))
        self.assertEqual(next(i1), 69)
        self.assertEqual(next(i1), 420)
        self.assertRaises(StopIteration, lambda: next(i1))
        self.assertRaises(StopIteration, lambda: next(iter(i1)))

        i2 = iter(Vector2Int(0, 69420))
        self.assertEqual(next(i2), 0)
        self.assertEqual(next(iter(i2)), 69420)
        self.assertRaises(StopIteration, lambda: next(iter(i2)))
        self.assertRaises(StopIteration, lambda: next(i2))

    def test_math(self):
        a = Vector2Int(3, 7)
        b = Vector2Int(5, 3)

        self.assertEqual(Vector2Int.distance(a, b), 4.47213595499957939282)
        self.assertEqual(Vector2Int.distance(Vector2Int(3, 0), Vector2Int(0, 2)), 3.60555127546398929312)

    def test_constructors1(self):
        half_even_down = Vector2(0.5, 0.5)
        half_even_up = Vector2(1.5, 1.5)
        half_even_updown = Vector2(69.5, 420.5)
        half_even_downup = Vector2(-69.5, -420.5)

        above_half_positive = Vector2(1.75, 1.75)
        above_half_negative = Vector2(-1.75, -1.75)

        below_half_positive = Vector2(1.25, 1.25)
        below_half_negative = Vector2(-1.25, -1.25)

        exact1 = Vector2(0.0, 0.0)
        exact2 = Vector2(69.0, 420.0)
        exact3 = Vector2(4.0, 69420.0)

        self.assertEqual(Vector2Int.ceil(half_even_down), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.ceil(half_even_up), Vector2Int(2, 2))
        self.assertEqual(Vector2Int.ceil(half_even_updown), Vector2Int(70, 421))
        self.assertEqual(Vector2Int.ceil(half_even_downup), Vector2Int(-69, -420))
        self.assertEqual(Vector2Int.ceil(above_half_positive), Vector2Int(2, 2))
        self.assertEqual(Vector2Int.ceil(above_half_negative), Vector2Int(-1, -1))
        self.assertEqual(Vector2Int.ceil(below_half_positive), Vector2Int(2, 2))
        self.assertEqual(Vector2Int.ceil(below_half_negative), Vector2Int(-1, -1))
        self.assertEqual(Vector2Int.ceil(exact1), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.ceil(exact2), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.ceil(exact3), Vector2Int(4, 69420))

        self.assertEqual(Vector2Int.floor(half_even_down), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.floor(half_even_up), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.floor(half_even_updown), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.floor(half_even_downup), Vector2Int(-70, -421))
        self.assertEqual(Vector2Int.floor(above_half_positive), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.floor(above_half_negative), Vector2Int(-2, -2))
        self.assertEqual(Vector2Int.floor(below_half_positive), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.floor(below_half_negative), Vector2Int(-2, -2))
        self.assertEqual(Vector2Int.floor(exact1), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.floor(exact2), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.floor(exact3), Vector2Int(4, 69420))

        self.assertEqual(Vector2Int.round(half_even_down), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.round(half_even_up), Vector2Int(2, 2))
        self.assertEqual(Vector2Int.round(half_even_updown), Vector2Int(70, 421))
        self.assertEqual(Vector2Int.round(half_even_downup), Vector2Int(-70, -421))
        self.assertEqual(Vector2Int.round(above_half_positive), Vector2Int(2, 2))
        self.assertEqual(Vector2Int.round(above_half_negative), Vector2Int(-2, -2))
        self.assertEqual(Vector2Int.round(below_half_positive), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.round(below_half_negative), Vector2Int(-1, -1))
        self.assertEqual(Vector2Int.round(exact1), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.round(exact2), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.round(exact3), Vector2Int(4, 69420))

        self.assertEqual(Vector2Int.trunc(half_even_down), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.trunc(half_even_up), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.trunc(half_even_updown), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.trunc(half_even_downup), Vector2Int(-69, -420))
        self.assertEqual(Vector2Int.trunc(above_half_positive), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.trunc(above_half_negative), Vector2Int(-1, -1))
        self.assertEqual(Vector2Int.trunc(below_half_positive), Vector2Int(1, 1))
        self.assertEqual(Vector2Int.trunc(below_half_negative), Vector2Int(-1, -1))
        self.assertEqual(Vector2Int.trunc(exact1), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.trunc(exact2), Vector2Int(69, 420))
        self.assertEqual(Vector2Int.trunc(exact3), Vector2Int(4, 69420))

    def test_constructors2(self):
        a = Vector2Int(2, 5)
        b1 = Vector2Int(3, 4)
        b2 = Vector2Int(1, 7)
        b3 = Vector2Int(0, 0)
        b4 = Vector2Int(10, 10)

        self.assertEqual(Vector2Int.min(a, b1), Vector2Int(2, 4))
        self.assertEqual(Vector2Int.min(a, b2), Vector2Int(1, 5))
        self.assertEqual(Vector2Int.min(a, b3), Vector2Int(0, 0))
        self.assertEqual(Vector2Int.min(a, b4), Vector2Int(2, 5))

        self.assertEqual(Vector2Int.max(a, b1), Vector2Int(3, 5))
        self.assertEqual(Vector2Int.max(a, b2), Vector2Int(2, 7))
        self.assertEqual(Vector2Int.max(a, b3), Vector2Int(2, 5))
        self.assertEqual(Vector2Int.max(a, b4), Vector2Int(10, 10))

    def test_converters(self):
        self.assertEqual(Vector2Int(3, 5).to_vector2(), Vector2(3.0, 5.0))
        self.assertEqual(Vector2Int(69, -420).to_vector2(), Vector2(69.0, -420.0))
        self.assertEqual(Vector2Int(-24, -42).to_vector2(), Vector2(-24.0, -42.0))

if __name__ == '__main__':
    unittest.main()



from enum import StrEnum as _StrEnum
import platform as _platform

class OSPlatform(_StrEnum):
    WINDOWS = "Windows"
    LINUX = "Linux"
    MAC_OS = "Darwin"
    JAVA = "Java"

def get_os_platform() -> OSPlatform:
    return OSPlatform(_platform.system())

def is_os_platform(os_platform: OSPlatform) -> bool:
    return _platform.system() == os_platform.value



import unittest

from nalpy.math import Vector2


class BasicFunctionality(unittest.TestCase):
    def test_index(self):
        v = Vector2(3.5, 7.25)
        self.assertEqual(v[0], 3.5)
        self.assertEqual(v[1], 7.25)
        self.assertRaises(IndexError, lambda: v[3])

        class MockIndexObj:
            def __init__(self, i: int) -> None:
                self.__internal_index = i
            def __index__(self) -> int:
                return self.__internal_index
        self.assertEqual(v[MockIndexObj(0)], 3.5)
        self.assertEqual(v[MockIndexObj(1)], 7.25)

    def test_xy(self):
        v = Vector2(3.5, 7.25)
        self.assertEqual(v.x, 3.5)
        self.assertEqual(v.y, 7.25)
        self.assertRaises(AttributeError, lambda: v.z) # type: ignore

    def test_equals(self):
        v = Vector2(3.0, 3.0)
        self.assertEqual(v, Vector2(3.0, 3.0))
        self.assertNotEqual(v, Vector2(2.0, 3.0))
        self.assertNotEqual(v, Vector2(3.0, 2.0))
        self.assertNotEqual(v, Vector2(2.0, 2.0))

        self.assertRaises(TypeError, lambda: v == (3.0, 3.0))
        self.assertNotEqual(v, None)
        self.assertNotEqual(None, v)

        self.assertEqual(Vector2(3, 3), Vector2(3.0, 3.0))
        self.assertEqual(Vector2(5, 2.5), Vector2(5.0, 2.5))
        self.assertNotEqual(Vector2(5, 2), Vector2(5.0, 2.5))

    def test_hash(self):
        test_values: tuple[tuple[float, float], ...] = (
            (69, 420),
            (420, 69),
            (69, 69),
            (420, 420),

            (69.420, 420.420),
            (420.420, 69.420),
            (69.69, 69.69),
            (420.420, 420.420),

            (0, 0),
            (0.0, 0.0),
            (-1, -1),
            (-69.420, -69.420),
            (1000, 1000),
            (1000.0, 1000.0),
            (-10000, -10000),
            (-10000.0, -10000.0),

            (float("inf"), float("inf")),
            (float("-inf"), float("-inf")),
            (float("inf"), float("-inf"))
        )

        # Vector2 hash should match tuple hash
        for v in test_values:
            self.assertEqual(hash(v), hash(Vector2(*v)), f"hash({v}) != hash({Vector2(*v)})")

        # Vector2 hash behaviour on nan values
        self.assertNotEqual(hash(Vector2(float("nan"), float("nan"))), hash(Vector2(float("nan"), float("nan"))))
        self.assertNotEqual(hash(Vector2(float("nan"), float("inf"))), hash(Vector2(float("nan"), float("inf"))))
        self.assertNotEqual(hash(Vector2(float("nan"), float("-inf"))), hash(Vector2(float("nan"), float("-inf"))))
        self.assertEqual(hash(Vector2(69.420, float("nan"))), hash(Vector2(69.420, float("nan"))))
        self.assertEqual(hash(Vector2(float("nan"), 69.420)), hash(Vector2(float("nan"), 69.420)))

    def test_constants(self):
        self.assertEqual(Vector2.zero,  Vector2(0.0, 0.0))
        self.assertEqual(Vector2.one,   Vector2(1.0, 1.0))
        self.assertEqual(Vector2.up,    Vector2(0.0, 1.0))
        self.assertEqual(Vector2.down,  Vector2(0.0, -1.0))
        self.assertEqual(Vector2.left,  Vector2(-1.0, 0.0))
        self.assertEqual(Vector2.right, Vector2(1.0, 0.0))

    def test_repr(self):
        self.assertEqual(repr(Vector2(2.0, 2.0)), f"Vector2(2.0, 2.0)")
        self.assertEqual(repr(Vector2(2.4, 3.075)), f"Vector2(2.4, 3.075)")

        self.assertEqual(repr(Vector2(2.0, 2.0)), str(Vector2(2.0, 2.0)))
        self.assertEqual(repr(Vector2(2.4, 3.075)), str(Vector2(2.4, 3.075)))

    def test_addsub(self):
        b: Vector2 = Vector2.one
        self.assertEqual(b + Vector2(2.5, 2.5), Vector2(3.5, 3.5))
        self.assertEqual(b + Vector2(0.0, 1.75), Vector2(1.0, 2.75))

        self.assertEqual(b - Vector2(2.5, 2.5), Vector2(-1.5, -1.5))
        self.assertEqual(b - Vector2(0.0, 1.75), Vector2(1.0, -0.75))

    def test_mult(self):
        b: Vector2 = Vector2.one
        self.assertEqual(b * 2, Vector2(2.0, 2.0))
        self.assertEqual(2 * b, Vector2(2.0, 2.0))
        self.assertEqual(b * 2.0, Vector2(2.0, 2.0))
        self.assertEqual(2.0 * b, Vector2(2.0, 2.0))

        self.assertEqual(Vector2(3, 4) * b, Vector2(3.0, 4.0))
        self.assertEqual(b * Vector2(3, 4), Vector2(3.0, 4.0))
        self.assertEqual(Vector2(3.0, 4.0) * b, Vector2(3.0, 4.0))
        self.assertEqual(b * Vector2(3.0, 4.0), Vector2(3.0, 4.0))

    def test_divmod(self):
        self.assertEqual(Vector2(9, 9.0) / 3, Vector2(3.0, 3))
        self.assertEqual(Vector2(9, 9.0) / 3.0, Vector2(3.0, 3))
        self.assertEqual(Vector2(9, 9.0) / Vector2(3, 6), Vector2(3.0, 1.5))
        self.assertEqual(Vector2(-9, -9.0) / 3, Vector2(-3.0, -3))
        self.assertEqual(Vector2(-9, -9.0) / 3.0, Vector2(-3.0, -3))
        self.assertEqual(Vector2(-9, -9.0) / Vector2(3, 6), Vector2(-3.0, -1.5))

        self.assertEqual(Vector2(10, 10.0) // 3, Vector2(3.0, 3))
        self.assertEqual(Vector2(10, 10.0) // 3.0, Vector2(3.0, 3))
        self.assertEqual(Vector2(10, 10.0) // Vector2(3, 6), Vector2(3.0, 1))
        self.assertEqual(Vector2(-10, -10.0) // 3, Vector2(-4.0, -4))
        self.assertEqual(Vector2(-10, -10.0) // 3.0, Vector2(-4.0, -4))
        self.assertEqual(Vector2(-10, -10.0) // Vector2(3, 6), Vector2(-4.0, -2))

        self.assertEqual(Vector2(10, 10.0) % 3, Vector2(1.0, 1))
        self.assertEqual(Vector2(10, 10.0) % Vector2(3, 5), Vector2(1.0, 0))
        self.assertEqual(Vector2(-10, -10.0) % 3, Vector2(2.0, 2))
        self.assertEqual(Vector2(-10, -10.0) % Vector2(3, 5), Vector2(2.0, 0.0))

        self.assertEqual(divmod(Vector2(10, 10.0), 3), (Vector2(3.0, 3), Vector2(1.0, 1)))
        self.assertEqual(divmod(Vector2(10, 10.0), Vector2(3, 5)), (Vector2(3.0, 2.0), Vector2(1.0, 0.0)))
        self.assertEqual(divmod(Vector2(-10, -10.0), 3), (Vector2(-4.0, -4), Vector2(2.0, 2)))
        self.assertEqual(divmod(Vector2(-10, -10.0), Vector2(3, 5)), (Vector2(-4.0, -2.0), Vector2(2.0, 0.0)))

    def test_rest_operators(self):
        self.assertEqual(-Vector2.one, Vector2(-1, -1))
        self.assertEqual(-Vector2.zero, Vector2.zero)
        self.assertEqual(-Vector2(-3, -3.0), Vector2(3.0, 3))

        self.assertEqual(abs(Vector2(-2, -2)), Vector2(2, 2))
        self.assertEqual(abs(Vector2(2, -2)), Vector2(2, 2))
        self.assertEqual(abs(Vector2(-2, 2)), Vector2(2, 2))
        self.assertEqual(abs(Vector2(2, 2)), Vector2(2, 2))

        self.assertEqual(abs(Vector2(-2.258, -2.258)), Vector2(2.258, 2.258))
        self.assertEqual(abs(Vector2(2.258, -2.258)), Vector2(2.258, 2.258))
        self.assertEqual(abs(Vector2(-2.258, 2.258)), Vector2(2.258, 2.258))
        self.assertEqual(abs(Vector2(2.258, 2.258)), Vector2(2.258, 2.258))

    def test_properties(self):
        self.assertEqual(Vector2(3, 6).magnitude, 6.70820393249936908923)
        self.assertEqual(Vector2(7, 12).magnitude, 13.89244398944980450843)
        self.assertEqual(Vector2(0.15, 0.75).magnitude, 0.7648529270389177245)

        unnormalized = Vector2(4.5, 3.42)
        self.assertEqual(unnormalized.magnitude, 5.65211464851872940138)
        normalized = unnormalized.normalized
        self.assertEqual(normalized.magnitude, 1.0)
        self.assertEqual(normalized.x, 0.79616219412310251879)
        self.assertEqual(normalized.y, 0.60508326753355791428)

    def test_iter(self):
        i1 = iter(Vector2(69.420, 420.69))
        self.assertEqual(next(i1), 69.420)
        self.assertEqual(next(i1), 420.69)
        self.assertRaises(StopIteration, lambda: next(i1))
        self.assertRaises(StopIteration, lambda: next(iter(i1)))

        i2 = iter(Vector2(0.0, float("inf")))
        self.assertEqual(next(i2), 0.0)
        self.assertEqual(next(iter(i2)), float("inf"))
        self.assertRaises(StopIteration, lambda: next(iter(i2)))
        self.assertRaises(StopIteration, lambda: next(i2))

    def test_math(self):
        a = Vector2(3.4, 7.6)
        a_perpendic = Vector2.perpendicular(a)

        b = Vector2(5.77, 3.22)

        self.assertEqual(Vector2.dot(a, b), 44.09)

        self.assertEqual(a_perpendic.x, -a.y)
        self.assertEqual(a_perpendic.y, a.x)
        self.assertEqual(Vector2.dot(a, a_perpendic), 0)

        self.assertAlmostEqual(Vector2.distance(a, b), 4.9800903606259997258)
        self.assertEqual(Vector2.distance(Vector2(3.0, 0.0), Vector2(0.0, 2.0)), 3.60555127546398929312)

    def test_angle(self):
        a = Vector2.right
        b1 = Vector2.up
        b2 = Vector2.down
        b3 = Vector2(1.0, -1.0)

        self.assertEqual(Vector2.angle(a, b1), 90)
        self.assertEqual(Vector2.angle(a, b2), 90)
        self.assertEqual(Vector2.signed_angle(a, b1), 90)
        self.assertEqual(Vector2.signed_angle(a, b2), -90)

        self.assertAlmostEqual(Vector2.angle(a, b3), 45)
        self.assertAlmostEqual(Vector2.signed_angle(a, b3), -45)

    def test_converters_and_constructors(self):
        a = Vector2(2.0, 5.0)
        b1 = Vector2(3.0, 4.0)
        b2 = Vector2(1.0, 7.0)
        b3 = Vector2(0.0, 0.0)
        b4 = Vector2(10, 10)

        self.assertEqual(Vector2.min(a, b1), Vector2(2.0, 4.0))
        self.assertEqual(Vector2.min(a, b2), Vector2(1.0, 5.0))
        self.assertEqual(Vector2.min(a, b3), Vector2(0.0, 0.0))
        self.assertEqual(Vector2.min(a, b4), Vector2(2.0, 5.0))

        self.assertEqual(Vector2.max(a, b1), Vector2(3.0, 5.0))
        self.assertEqual(Vector2.max(a, b2), Vector2(2.0, 7.0))
        self.assertEqual(Vector2.max(a, b3), Vector2(2.0, 5.0))
        self.assertEqual(Vector2.max(a, b4), Vector2(10.0, 10.0))

    def test_interpolation(self):
        a = Vector2(2, 4)
        b = Vector2(14.5, 7.85)

        lerp75 = Vector2.lerp(a, b, 0.75)
        lerpuncl75 = Vector2.lerp_unclamped(a, b, 0.75)
        res75 = Vector2(11.375, 6.8875)

        self.assertEqual(Vector2.lerp(a, b, 0.5), Vector2(8.25, 5.925))
        self.assertEqual(Vector2.lerp(a, b, 0.25), Vector2(5.125, 4.9625))
        self.assertAlmostEqual(lerp75.x, res75.x)
        self.assertAlmostEqual(lerp75.y, res75.y)
        self.assertEqual(Vector2.lerp(a, b, 0.0), a)
        self.assertEqual(Vector2.lerp(a, b, -10.0), a)
        self.assertEqual(Vector2.lerp(a, b, 1.0), b)
        self.assertEqual(Vector2.lerp(a, b, 10.0), b)

        self.assertEqual(Vector2.lerp_unclamped(a, b, 0.5), Vector2(8.25, 5.925))
        self.assertEqual(Vector2.lerp_unclamped(a, b, 0.25), Vector2(5.125, 4.9625))
        self.assertAlmostEqual(lerpuncl75.x, res75.x)
        self.assertAlmostEqual(lerpuncl75.y, res75.y)
        self.assertEqual(Vector2.lerp_unclamped(a, b, 0.0), a)
        self.assertEqual(Vector2.lerp_unclamped(a, b, -10.0), Vector2(-123, -34.5))
        self.assertEqual(Vector2.lerp_unclamped(a, b, 1.0), b)
        self.assertEqual(Vector2.lerp_unclamped(a, b, 10.0), Vector2(127, 42.5))

        # move_towards checks are generated.
        self.assertEqual(Vector2.move_towards(Vector2(29.431186908718228, 33.6621820548442), Vector2(23.120215968907857, 59.38576299316706), 66.61514209773944), Vector2(23.120215968907857, 59.38576299316706))
        self.assertEqual(Vector2.move_towards(Vector2(33.79504419784424, 15.245959830973579), Vector2(29.8970235981436, 56.87761092723505), 45.60931683390097), Vector2(29.8970235981436, 56.87761092723505))
        self.assertEqual(Vector2.move_towards(Vector2(36.967494011692054, 48.05636806402712), Vector2(64.7026491150305, 10.799301027132651), 22.62248852523597), Vector2(50.476173519781156, 29.90994789303563))
        self.assertEqual(Vector2.move_towards(Vector2(36.967494011692054, 48.05636806402712), Vector2(64.7026491150305, 10.799301027132651), 22.62248852523597), Vector2(50.476173519781156, 29.90994789303563))
        self.assertEqual(Vector2.move_towards(Vector2(17.304290257053363, 70.00883344211574), Vector2(25.30328000212717, 62.156384285746164), 73.68637824328094), Vector2(25.30328000212717, 62.156384285746164))
        self.assertEqual(Vector2.move_towards(Vector2(7.345439149536899, 27.733716913826783), Vector2(43.81449326240699, 53.65251153801417), 0.53), Vector2(7.7774480048922054, 28.04074842506724))

        # Max delta over the moveable distance
        self.assertEqual(Vector2.move_towards(Vector2(8.657950152375955, 4.180694783383892), Vector2(60.89320793752563, 64.02011771066626), 84948.51669138268), Vector2(60.89320793752563, 64.02011771066626))

if __name__ == '__main__':
    unittest.main()



#cython: language_level=3

from libc.math cimport hypot, ceil, floor, llround
from libc.stdlib cimport llabs

from .vector2 cimport Vector2

cdef extern from "Python.h":
    int SIZEOF_PY_HASH_T

# Hashing
ctypedef unsigned long long int _Vec_uhash_t

cdef _Vec_uhash_t _VecHASH_XXPRIME_1 = 11400714785074694791ULL
cdef _Vec_uhash_t _VecHASH_XXPRIME_2 = 14029467366897019727ULL
cdef _Vec_uhash_t _VecHASH_XXPRIME_5 = 2870177450012600261ULL
cdef inline _Vec_uhash_t _VecHASH_XXROTATE(_Vec_uhash_t x):
    return ((x << 31) | (x >> 33)) # Rotate left 31 bits
# Hashing

cdef class Vector2Int:
    zero = Vector2Int(0, 0)
    one = Vector2Int(1, 1)

    up = Vector2Int(0, 1)
    down = Vector2Int(0, -1)
    left = Vector2Int(-1, 0)
    right = Vector2Int(1, 0)


    def __init__(self, _V2I_int_t x, _V2I_int_t y):
        # Vector2Int(x, y)
        self.x = x
        self.y = y

    def __getitem__(self, Py_ssize_t i):
        # self[i]
        if i == 0:
            return self.x
        if i == 1:
            return self.y

        raise IndexError(i)

    def __repr__(self):
        # repr(self)
        return f"Vector2Int({self.x}, {self.y})"

    def __len__(self):
        # len(self)
        return 2

    def __add__(self, Vector2Int other):
        # self + other
        return Vector2Int(self.x + other.x, self.y + other.y)

    def __sub__(self, Vector2Int other):
        # self - other
        return Vector2Int(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        # self * other
        cdef _V2I_int_t x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef _V2I_int_t y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        return Vector2Int(self.x * x, self.y * y)

    def __rmul__(self, other):
        # other * self
        # Duplicated code because inline function didn't work for some reason...
        cdef _V2I_int_t x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef _V2I_int_t y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        return Vector2Int(self.x * x, self.y * y)

    def __truediv__(self, other):
        # self / other, NOTE: Returns Vector2
        cdef double x
        cdef double y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x / x, self.y / y)

    def __floordiv__(self, other):
        # self // other
        cdef _V2I_int_t x
        cdef _V2I_int_t y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        return Vector2Int(self.x // x, self.y // y)

    def __mod__(self, other):
        # self % other
        cdef _V2I_int_t x
        cdef _V2I_int_t y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        return Vector2Int(self.x % x, self.y % y)

    def __divmod__(self, other):
        # divmod(self, other)
        cdef _V2I_int_t x
        cdef _V2I_int_t y
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        cdef _V2I_int_t x_fdiv = self.x // x # Cython doesn't have a divmod variant for C
        cdef _V2I_int_t x_mod = self.x % x   # We calculate these separately to avoid using Python's divmod
        cdef _V2I_int_t y_fdiv = self.y // y
        cdef _V2I_int_t y_mod = self.y % y
        return (Vector2Int(x_fdiv, y_fdiv), Vector2Int(x_mod, y_mod))

    def __neg__(self):
        # -self
        return Vector2Int(-self.x, -self.y)

    def __abs__(self):
        # abs(self)
        return Vector2Int(llabs(self.x), llabs(self.y))

    def __eq__(self, Vector2Int other):
        # self == other
        return self.x == other.x and self.y == other.y

    def __iter__(self):
        # iter(self)
        return Vector2IntIterator(self)

    # Adapted from tuplehash https://github.com/python/cpython/blob/3.11/Objects/tupleobject.c#L321
    # Doesn't work when extracted into a .pxd file for some reason, that's why this has been copied from vector2.pyx
    def __hash__(self):
        if SIZEOF_PY_HASH_T != 8:
            raise RuntimeError("64 bit hash type required.")

        cdef _Vec_uhash_t xlane = <_Vec_uhash_t>hash(self.x)
        cdef _Vec_uhash_t ylane = <_Vec_uhash_t>hash(self.y)

        if xlane == <_Vec_uhash_t>-1 or ylane == <_Vec_uhash_t>-1:
            return -1

        cdef _Vec_uhash_t acc = _VecHASH_XXPRIME_5

        # X
        acc += xlane * _VecHASH_XXPRIME_2
        acc = _VecHASH_XXROTATE(acc)
        acc *= _VecHASH_XXPRIME_1

        # Y
        acc += ylane * _VecHASH_XXPRIME_2
        acc = _VecHASH_XXROTATE(acc)
        acc *= _VecHASH_XXPRIME_1

        acc += (<Py_ssize_t>2) ^ (_VecHASH_XXPRIME_5 ^ 3527539UL)
        # To keep compatibility with tuple's hash implementation
        # The performance improvement by removing this is negligible

        if acc == <_Vec_uhash_t>-1:
            return 1546275796

        return acc

    @property
    def magnitude(self):
        return hypot(<double>self.x, <double>self.y)

    @staticmethod
    def distance(Vector2Int a, Vector2Int b):
        cdef _V2I_int_t diff_x = a.x - b.x
        cdef _V2I_int_t diff_y = a.y - b.y
        return hypot(<double>diff_x, <double>diff_y)

    @staticmethod
    def ceil(v: Vector2) -> Vector2Int:
       return Vector2Int(<_V2I_int_t>ceil(v.x), <_V2I_int_t>ceil(v.y))

    @staticmethod
    def floor(v: Vector2) -> Vector2Int:
        return Vector2Int(<_V2I_int_t>floor(v.x), <_V2I_int_t>floor(v.y))

    @staticmethod
    def round(v: Vector2) -> Vector2Int:
        return Vector2Int(<_V2I_int_t>llround(v.x), <_V2I_int_t>llround(v.y))
        # Casting in case we change the _V2I_int_t later. long long int can always hold a rounded double.

    @staticmethod
    def trunc(v: Vector2) -> Vector2Int:
        return Vector2Int(<_V2I_int_t>v.x, <_V2I_int_t>v.y) # Casting floating point to integer truncates

    @staticmethod
    def min(Vector2Int a, Vector2Int b):
        return Vector2Int(min(a.x, b.x), min(a.y, b.y))

    @staticmethod
    def max(Vector2Int a, Vector2Int b):
        return Vector2Int(max(a.x, b.x), max(a.y, b.y))

    def to_vector2(self) -> Vector2:
        return Vector2(<double>self.x, <double>self.y)

    def to_tuple(self):
        return (self.x, self.y)

    def to_dict(self):
        return {"x": self.x, "y": self.y}

cdef class Vector2IntIterator:
    cdef Vector2Int vec2
    cdef char index

    def __init__(self, Vector2Int vec2):
        self.vec2 = vec2
        self.index = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.index += 1
        if self.index == 0:
            return self.vec2.x
        if self.index == 1:
            return self.vec2.y
        raise StopIteration()



import re

class _Rule:
    def __init__(self, pattern: str, replacement: str) -> None:
        self._regex = re.compile(pattern, re.IGNORECASE)
        self._replacement = replacement

    def apply(self, word: str) -> str | None:
        if self._regex.search(word) is None:
            return None

        return self._regex.sub(self._replacement, word)

class Vocabulary:
    def __init__(self) -> None:
        self._plurals: list[_Rule] = []
        self._singulars: list[_Rule] = []
        self._uncountables: list[str] = []

    def add_irregular(self, singular: str, plural: str, match_ending: bool = True):
        if match_ending:
            self.add_plural("(" + singular[0] + ")" + singular[1:] + "$", "\\1" + plural[1:])
            self.add_singular("(" + plural[0] + ")" + plural[1:] + "$", "\\1" + singular[1:])
        else:
            self.add_plural(f"^{singular}$", plural)
            self.add_singular(f"^{plural}$", singular)

    def add_uncountable(self, word: str):
        self._uncountables.append(word.lower())

    def add_plural(self, rule: str, replacement: str):
        self._plurals.append(_Rule(rule, replacement))

    def add_singular(self, rule: str, replacement: str):
        self._singulars.append(_Rule(rule, replacement))

    def pluralize(self, word: str, is_known_to_be_singular: bool = True) -> str:
        result = self._apply_rules(self._plurals, word, False)

        if is_known_to_be_singular:
            return result if result is not None else word

        # the singularity is unknown so we should check all possibilities
        as_singular = self._apply_rules(self._singulars, word, False)
        as_singular_as_plural = self._apply_rules(self._plurals, as_singular, False)
        if (
            as_singular is not None
            and as_singular != word
            and as_singular + "s" != word
            and as_singular_as_plural == word
            and result != word
        ):
            return word

        assert result is not None
        return result

    def singularize(self, word: str, is_known_to_be_plural: bool = True, skip_simple_words: bool = False) -> str:
        result = self._apply_rules(self._singulars, word, skip_simple_words)

        if is_known_to_be_plural:
            return result if result is not None else word

        # the plurality is unknown so we should check all possibilities
        as_plural = self._apply_rules(self._plurals, word, False)
        as_plural_as_singular = self._apply_rules(self._singulars, as_plural, False)
        if (
            as_plural != word
            and word + "s" != as_plural
            and as_plural_as_singular == word
            and result != word
        ):
            return word

        return result if result is not None else word

    def _apply_rules(self, rules: list[_Rule], word: str | None, skip_first_rule: bool) -> str | None:
        if word is None:
            return None

        if len(word) < 1:
            return word

        if self._is_uncountable(word):
            return word

        result = word
        for rule in reversed(rules[(1 if skip_first_rule else 0):]):
            result = rule.apply(word)
            if result != None:
                break

        return self._match_upper_case(word, result) if result is not None else None

    def _is_uncountable(self, word: str):
        return word.lower() in self._uncountables

    def _match_upper_case(self, word: str, replacement: str):
        return (replacement[0].upper() + replacement[1:]) if word[0].isupper() and replacement[0].islower() else replacement


def _generate_default_vocabulary() -> Vocabulary:
    default: Vocabulary = Vocabulary()

    default.add_plural("$", "s")
    default.add_plural("s$", "s")
    default.add_plural("(ax|test)is$", "\\1es")
    default.add_plural("(octop|vir|alumn|fung|cact|foc|hippopotam|radi|stimul|syllab|nucle)us$", "\\1i")
    default.add_plural("(alias|bias|iris|status|campus|apparatus|virus|walrus|trellis)$", "\\1es")
    default.add_plural("(buffal|tomat|volcan|ech|embarg|her|mosquit|potat|torped|vet)o$", "\\1oes")
    default.add_plural("([dti])um$", "\\1a")
    default.add_plural("sis$", "ses")
    default.add_plural("(?:([^f])fe|([lr])f)$", "\\1\\2ves")
    default.add_plural("(hive)$", "\\1s")
    default.add_plural("([^aeiouy]|qu)y$", "\\1ies")
    default.add_plural("(x|ch|ss|sh)$", "\\1es")
    default.add_plural("(matr|vert|ind|d)(ix|ex)$", "\\1ices")
    default.add_plural("(^[m|l])ouse$", "\\1ice")
    default.add_plural("^(ox)$", "\\1en")
    default.add_plural("(quiz)$", "\\1zes")
    default.add_plural("(buz|blit|walt)z$", "\\1zes")
    default.add_plural("(hoo|lea|loa|thie)f$", "\\1ves")
    default.add_plural("(alumn|alg|larv|vertebr)a$", "\\1ae")
    default.add_plural("(criteri|phenomen)on$", "\\1a")

    default.add_singular("s$", "")
    default.add_singular("(n)ews$", "\\1ews")
    default.add_singular("([dti])a$", "\\1um")
    default.add_singular("(analy|ba|diagno|parenthe|progno|synop|the|ellip|empha|neuro|oa|paraly)ses$", "\\1sis")
    default.add_singular("([^f])ves$", "\\1fe")
    default.add_singular("(hive)s$", "\\1")
    default.add_singular("(tive)s$", "\\1")
    default.add_singular("([lr]|hoo|lea|loa|thie)ves$", "\\1f")
    default.add_singular("(^zomb)?([^aeiouy]|qu)ies$", "\\2y")
    default.add_singular("(s)eries$", "\\1eries")
    default.add_singular("(m)ovies$", "\\1ovie")
    default.add_singular("(x|ch|ss|sh)es$", "\\1")
    default.add_singular("(^[m|l])ice$", "\\1ouse")
    default.add_singular("(?<!^[a-z])(o)es$", "\\1")
    default.add_singular("(shoe)s$", "\\1")
    default.add_singular("(cris|ax|test)es$", "\\1is")
    default.add_singular("(octop|vir|alumn|fung|cact|foc|hippopotam|radi|stimul|syllab|nucle)i$", "\\1us")
    default.add_singular("(alias|bias|iris|status|campus|apparatus|virus|walrus|trellis)es$", "\\1")
    default.add_singular("^(ox)en", "\\1")
    default.add_singular("(matr|d)ices$", "\\1ix")
    default.add_singular("(vert|ind)ices$", "\\1ex")
    default.add_singular("(quiz)zes$", "\\1")
    default.add_singular("(buz|blit|walt)zes$", "\\1z")
    default.add_singular("(alumn|alg|larv|vertebr)ae$", "\\1a")
    default.add_singular("(criteri|phenomen)a$", "\\1on")
    default.add_singular("([b|r|c]ook|room|smooth)ies$", "\\1ie")

    default.add_irregular("person", "people")
    default.add_irregular("man", "men")
    default.add_irregular("human", "humans")
    default.add_irregular("child", "children")
    default.add_irregular("sex", "sexes")
    default.add_irregular("glove", "gloves")
    default.add_irregular("move", "moves")
    default.add_irregular("goose", "geese")
    default.add_irregular("wave", "waves")
    default.add_irregular("foot", "feet")
    default.add_irregular("tooth", "teeth")
    default.add_irregular("curriculum", "curricula")
    default.add_irregular("database", "databases")
    default.add_irregular("zombie", "zombies")
    default.add_irregular("personnel", "personnel")
    default.add_irregular("cache", "caches")

    default.add_irregular("ex", "exes", match_ending=False)
    default.add_irregular("is", "are", match_ending=False)
    default.add_irregular("that", "those", match_ending=False)
    default.add_irregular("this", "these", match_ending=False)
    default.add_irregular("bus", "buses", match_ending=False)
    default.add_irregular("die", "dice", match_ending=False)
    default.add_irregular("tie", "ties", match_ending=False)

    default.add_uncountable("staff")
    default.add_uncountable("training")
    default.add_uncountable("equipment")
    default.add_uncountable("information")
    default.add_uncountable("corn")
    default.add_uncountable("milk")
    default.add_uncountable("rice")
    default.add_uncountable("money")
    default.add_uncountable("species")
    default.add_uncountable("series")
    default.add_uncountable("fish")
    default.add_uncountable("sheep")
    default.add_uncountable("deer")
    default.add_uncountable("aircraft")
    default.add_uncountable("oz")
    default.add_uncountable("tsp")
    default.add_uncountable("tbsp")
    default.add_uncountable("ml")
    default.add_uncountable("l")
    default.add_uncountable("water")
    default.add_uncountable("waters")
    default.add_uncountable("semen")
    default.add_uncountable("sperm")
    default.add_uncountable("bison")
    default.add_uncountable("grass")
    default.add_uncountable("hair")
    default.add_uncountable("mud")
    default.add_uncountable("elk")
    default.add_uncountable("luggage")
    default.add_uncountable("moose")
    default.add_uncountable("offspring")
    default.add_uncountable("salmon")
    default.add_uncountable("shrimp")
    default.add_uncountable("someone")
    default.add_uncountable("swine")
    default.add_uncountable("trout")
    default.add_uncountable("tuna")
    default.add_uncountable("corps")
    default.add_uncountable("scissors")
    default.add_uncountable("means")
    default.add_uncountable("mail")
    default.add_uncountable("metadata")

    return default

_default_vocabulary: Vocabulary | None = None
def default() -> Vocabulary:
    global _default_vocabulary
    if _default_vocabulary is None:
        _default_vocabulary = _generate_default_vocabulary()
    return _default_vocabulary



#cython: language_level=3

from libc.math cimport hypot

from .vector2_int cimport Vector2Int

ctypedef long long int int_t

# Modified from MVector2
cdef class MVector2Int:
    @staticmethod
    def zero():
        return MVector2Int(0, 0)
    @staticmethod
    def one():
        return MVector2Int(1, 1)
    @staticmethod
    def up():
        return MVector2Int(0, 1)
    @staticmethod
    def down():
        return MVector2Int(0, -1)
    @staticmethod
    def left():
        return MVector2Int(-1, 0)
    @staticmethod
    def right():
        return MVector2Int(1, 0)

    cdef public int_t x
    cdef public int_t y

    def __init__(self, int_t x, int_t y):
        self.x = x
        self.y = y

    @staticmethod
    def from_immutable(Vector2Int immutable):
        return MVector2Int(immutable.x, immutable.y)

    def __getitem__(self, Py_ssize_t i):
        if i == 0:
            return self.x
        if i == 1:
            return self.y

        raise IndexError(i)

    def __repr__(self):
        return f"MVector2Int({self.x}, {self.y})"

    def __len__(self):
        return 2

    # Only in-place arithmetic supported. For other arithmetic operations, conversion to Vector2Int is required.
    def __iadd__(self, other):
        if not isinstance(other, (MVector2Int, Vector2Int)):
            return NotImplemented

        self.x += <int_t>other.x # Casting to force C addition instead of Python addition
        self.y += <int_t>other.y

        return self

    def __isub__(self, other):
        if not isinstance(other, (MVector2Int, Vector2Int)):
            return NotImplemented

        self.x += <int_t>other.x
        self.y += <int_t>other.y

        return self

    def __imul__(self, other):
        cdef int_t x
        cdef int_t y
        if isinstance(other, (MVector2Int, Vector2Int)):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        self.x *= x
        self.y *= y

        return self

    def __ifloordiv__(self, other):
        cdef int_t x
        cdef int_t y
        if isinstance(other, (MVector2Int, Vector2Int)):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        self.x //= x
        self.y //= y

        return self

    def __imod__(self, other):
        cdef int_t x
        cdef int_t y
        if isinstance(other, (MVector2Int, Vector2Int)):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = y = other
        else:
            return NotImplemented

        self.x %= x
        self.y %= y

        return self

    def __eq__(self, MVector2Int other):
        return self.x == other.x and self.y == other.y

    # NOTE: Object is mutable, do not implement __hash__


    @property
    def magnitude(self):
        return hypot(<double>self.x, <double>self.y)

    def copy(self):
        return MVector2Int(self.x, self.y)


    @staticmethod
    def min(MVector2Int a, MVector2Int b):
        return MVector2Int(min(a.x, b.x), min(a.y, b.y))

    @staticmethod
    def max(MVector2Int a, MVector2Int b):
        return MVector2Int(max(a.x, b.x), max(a.y, b.y))


    def to_immutable(self):
        return Vector2Int(self.x, self.y)

    def to_tuple(self):
        return (self.x, self.y)

    def to_dict(self):
        return {"x": self.x, "y": self.y}



from time import time as _time
import typing as _typing
from nalpy import console_utils as _console_utils

class SpinnerStyle(_typing.NamedTuple):
    frames: _typing.Sequence[str] = ('◷', '◶', '◵', '◴')
    frame_length: float = 0.25
    hide_cursor: bool = True

    @classmethod
    @property
    def default(cls) -> _typing.Self:
        return SpinnerStyle()

class Spinner:
    def __init__(self, style: SpinnerStyle) -> None:
        self._style: SpinnerStyle = style
        self._starttime: float | None = None
        self._cursor_hidden: bool = False

    def __enter__(self) -> _typing.Self:
        self.start()
        return self

    def __exit__(self, *_) -> bool:
        self.stop()
        return False

    def start(self):
        if self._style.hide_cursor:
            _console_utils.cursor_hide()
            self._cursor_hidden = True
        self._starttime = _time()

    def update(self, message: str) -> float:
        """Updates the spinner

        Returns:
            float: The elapsed seconds
        """
        if self._starttime is None:
            raise RuntimeError("Spinner has to be started before updating!")

        elapsed_seconds = _time() - self._starttime
        frame_index: int = int(elapsed_seconds // self._style.frame_length)
        frame_index %= len(self._style.frames)

        _console_utils.carriage_return()
        print(self._style.frames[frame_index], end=" ")
        print(message, end="")

        return elapsed_seconds

    def stop(self) -> None:
        """Resets the cursor visibility if it was modified."""
        if self._cursor_hidden:
            _console_utils.cursor_show()
            self._cursor_hidden = False



from nalpy.humanizer.numbers.enums import NumberFormat as NumberFormat

from nalpy.humanizer.numbers.heading import to_heading as to_heading
from nalpy.humanizer.numbers.heading import to_abbreviated_heading as to_abbreviated_heading
from nalpy.humanizer.numbers.heading import to_heading_arrow as to_heading_arrow

from nalpy.humanizer.numbers.count import ordinalize as ordinalize
from nalpy.humanizer.numbers.count import to_spaced_numerals as to_spaced_numerals
from nalpy.humanizer.numbers.count import to_words as to_words
from nalpy.humanizer.numbers.count import to_ordinal_words as to_ordinal_words
from nalpy.humanizer.numbers.count import to_roman as to_roman
from nalpy.humanizer.numbers.count import to_percent as to_percent
from nalpy.humanizer.numbers.count import tupleize as tupleize

from nalpy.humanizer.numbers.metric import MetricPrefix as MetricPrefix
from nalpy.humanizer.numbers.metric import BinaryPrefix as BinaryPrefix
from nalpy.humanizer.numbers.metric import convert_metric as convert_metric
from nalpy.humanizer.numbers.metric import to_metric as to_metric
from nalpy.humanizer.numbers.metric import to_short_metric as to_short_metric



units_map: tuple[str, ...] = ( "zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen", "seventeen", "eighteen", "nineteen" )
tens_map: tuple[str, ...] = ( "zero", "ten", "twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety" )
ordinal_exceptions: dict[int, str] = {
    1: "first",
    2: "second",
    3: "third",
    4: "fourth",
    5: "fifth",
    8: "eighth",
    9: "ninth",
    12: "twelfth"
}

def convert(number: int, is_ordinal: bool = False, add_and: bool = True) -> str:
    if number == 0:
        return _get_unit_value(0, is_ordinal)

    if number < 0:
        return f"minus {convert(-number)}"

    parts: list[str] = []

    if int(number / 1_000_000_000_000_000_000) > 0:
        parts.append(f"{convert(int(number / 1_000_000_000_000_000_000))} quintillion")
        number %= 1_000_000_000_000_000_000

    if int(number / 1_000_000_000_000_000) > 0:
        parts.append(f"{convert(int(number / 1_000_000_000_000_000))} quadrillion")
        number %= 1_000_000_000_000_000

    if int(number / 1_000_000_000_000) > 0:
        parts.append(f"{convert(int(number / 1_000_000_000_000))} trillion")
        number %= 1_000_000_000_000

    if int(number / 1000000000) > 0:
        parts.append(f"{convert(int(number / 1000000000))} billion")
        number %= 1000000000

    if int(number / 1000000) > 0:
        parts.append(f"{convert(int(number / 1000000))} million")
        number %= 1000000

    if int(number / 1000) > 0:
        parts.append(f"{convert(int(number / 1000))} thousand")
        number %= 1000

    if int(number / 100) > 0:
        parts.append(f"{convert(int(number / 100))} hundred")
        number %= 100

    if number > 0:
        if len(parts) > 0 and add_and:
            parts.append("and")

        if number < 20:
            parts.append(_get_unit_value(number, is_ordinal))
        else:
            last_part = tens_map[int(number / 10)]
            if (number % 10) > 0:
                last_part += f"-{_get_unit_value(number % 10, is_ordinal)}"
            elif is_ordinal:
                last_part = last_part.rstrip("y") + "ieth"

            parts.append(last_part)
    elif is_ordinal:
        parts[-1] += "th"

    to_words = " ".join(parts)

    if is_ordinal:
        to_words = _remove_one_prefix(to_words)

    return to_words



def _get_unit_value(number: int, is_ordinal: bool) -> str:
    if is_ordinal:
        if number in ordinal_exceptions:
            return ordinal_exceptions[number]

        return units_map[number] + "th"

    return units_map[number]

def _remove_one_prefix(to_words: str) -> str:
    if to_words.casefold().startswith("one".casefold()):
        return to_words[4:]
    return to_words



from typing import Self, SupportsIndex
from .vector2_int import Vector2Int

class MVector2Int:
    @staticmethod
    def zero() -> MVector2Int: ...
    @staticmethod
    def one() -> MVector2Int: ...
    @staticmethod
    def up() -> MVector2Int: ...
    @staticmethod
    def down() -> MVector2Int: ...
    @staticmethod
    def left() -> MVector2Int: ...
    @staticmethod
    def right() -> MVector2Int: ...

    x: int
    y: int

    def __init__(self, x: int, y: int) -> None: ...

    @staticmethod
    def from_immutable(immutable: Vector2Int) -> MVector2Int: ...

    def __getitem__(self, i: SupportsIndex) -> int: ...

    def __repr__(self) -> str: ...

    def __len__(self) -> int: ...

    def __iadd__(self, other: MVector2Int | Vector2Int) -> Self: ...

    def __isub__(self, other: MVector2Int | Vector2Int) -> Self: ...

    def __imul__(self, other: MVector2Int | Vector2Int | int) -> Self: ...

    def __ifloordiv__(self, other: MVector2Int | Vector2Int | int) -> Self: ...

    def __imod__(self, other: MVector2Int | Vector2Int | int) -> Self: ...

    def __eq__(self, other: MVector2Int) -> bool: ...

    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        ...

    def copy(self) -> MVector2Int:
        """Make a new vector where the components' values match this vectors values."""
        ...

    @staticmethod
    def min(a: MVector2Int, b: MVector2Int) -> MVector2Int:
        """Create a new `MVector2Int` by selecting the smallest components of the two given `MVector2Int` instances."""
        ...

    @staticmethod
    def max(a: MVector2Int, b: MVector2Int) -> MVector2Int:
        """Create a new `MVector2Int` by selecting the largest components of the two given `MVector2Int` instances."""
        ...

    def to_immutable(self) -> Vector2Int: ...
    def to_tuple(self) -> tuple[int, int]: ...
    def to_dict(self) -> dict[str, int]: ...



from nalpy.console_utils import ConsoleColor, ConsoleStyle
import nalpy.console_utils._helpers as _helper

# NOTE: Remember to add to console_utils public imports

def set_foreground_color(color: ConsoleColor):
    ansi = _helper.get_ansi_color(color, prefix_number=3)
    _helper.set_sgr(ansi)

def set_background_color(color: ConsoleColor): # NOTE: This leaves a long colored line in VSCode for some reason
    ansi = _helper.get_ansi_color(color, prefix_number=4)
    _helper.set_sgr(ansi)

def set_style(style: ConsoleStyle):
    _helper.set_sgr(str(style))

def reset_attributes():
    _helper.set_sgr("0")



#cython: language_level=3

cdef inline char doublesign(double x) noexcept: # Extracted into a separate function because cpdef doesn't support position only arguments.
    return (<unsigned char>(x > 0.0)) - (<unsigned char>(x < 0.0))



from typing import Sequence


def to_heading(degrees: float) -> str:
    return _get_heading_from_set(degrees, _HEADINGS_LONG)

def to_abbreviated_heading(degrees: float) -> str:
    return _get_heading_from_set(degrees, _HEADINGS_SHORT)

def to_heading_arrow(degrees: float) -> str:
    return _get_heading_from_set(degrees, _HEADINGS_ARROW)


_HEADINGS_LONG = ( "north", "north-northeast", "northeast", "east-northeast", "east", "east-southeast", "southeast", "south-southeast", "south", "south-southwest", "southwest", "west-southwest", "west", "west-northwest", "northwest", "north-northwest" )
_HEADINGS_SHORT = ( "N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE", "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW" )
_HEADINGS_ARROW = ( '↑', '↗', '→', '↘', '↓', '↙', '←', '↖' )

def _get_heading_from_set(heading: float, collection: Sequence[str]):
    ln = len(collection)
    val = round(heading / (360 / ln))
    return collection[val % ln]



_ROMAN_NUMERALS: dict[str, int] = {
    "M":  1000,
    "CM": 900,
    "D":  500,
    "CD": 400,
    "C":  100,
    "XC": 90,
    "L":  50,
    "XL": 40,
    "X":  10,
    "IX": 9,
    "V":  5,
    "IV": 4,
    "I":  1
}

def to_roman(num: int) -> str:
    if not (1 <= num <= 3999):
        raise ValueError("Number out of range.")

    output = ""
    for letters, value in _ROMAN_NUMERALS.items():
        while (num // value) > 0:
            output += letters
            num -= value

    return output



import typing as _typing

class _Helpers:
    @staticmethod
    def parse_pascal_or_camel(pascal_or_camel: str) -> str:
        def generator() -> _typing.Iterable[str]:
            characters = list(pascal_or_camel)
            for i in range(len(characters)):
                if i == 0:
                    yield characters[i].upper()
                elif characters[i].isupper() and not characters[i - 1].isupper():
                    yield " "
                    if i + 1 >= len(characters) or not characters[i + 1].isupper():
                        yield characters[i].lower()
                    else:
                        yield characters[i]
                else:
                    yield characters[i]

        return "".join(generator())

    @staticmethod
    def generate_pascal_or_camel(sentence: str, first_is_upper: bool):
        words = sentence.strip(" ").split(" ")

        if len(words) > 0:
            first_char = words[0][0].lower()
            if first_is_upper:
                first_char = first_char.upper()
            words[0] = first_char + words[0][1:]

        for i in range(1, len(words)):
            words[i] = words[i].capitalize() if not words[i].isupper() else words[i]
        return "".join(words)


    @staticmethod
    def parse_snake_or_kebab(snake_or_kebab: str, delimiter: str):
        low = " ".join(snake_or_kebab.split(delimiter))
        return low[0].upper() + low[1:]

    @staticmethod
    def generate_snake_or_kebab(sentence: str, delimiter: str):
        return delimiter.join(sentence.lower().split(" "))


def pascal_case_to_sentence(pascal: str) -> str:
    if len(pascal) < 1:
        return ""

    if " " in pascal or not pascal[0].isupper():
        raise ValueError("Not PascalCase!")

    return "".join(_Helpers.parse_pascal_or_camel(pascal))

def sentence_to_pascal_case(sentence: str) -> str:
    return _Helpers.generate_pascal_or_camel(sentence, True)


def camel_case_to_sentence(camel: str) -> str:
    if len(camel) < 1:
        return ""

    if " " in camel or not camel[0].islower():
        raise ValueError("Not camelCase!")

    return "".join(_Helpers.parse_pascal_or_camel(camel))

def sentence_to_camel_case(sentence: str) -> str:
    return _Helpers.generate_pascal_or_camel(sentence, False)


def snake_case_to_sentence(snake: str) -> str:
    if not snake.islower() or " " in snake:
        raise ValueError("Not snake_case!")

    return _Helpers.parse_snake_or_kebab(snake, "_")

def sentence_to_snake_case(sentence: str) -> str:
    return _Helpers.generate_snake_or_kebab(sentence, "_")


def kebab_case_to_sentence(kebab: str) -> str:
    if not kebab.islower() or " " in kebab:
        raise ValueError("Not kebab-case!")

    return _Helpers.parse_snake_or_kebab(kebab, "-")

def sentence_to_kebab_case(sentence: str) -> str:
    return _Helpers.generate_snake_or_kebab(sentence, "-")


def upper_snake_case_to_sentence(upper_snake: str) -> str:
    if not upper_snake.isupper() or " " in upper_snake:
        raise ValueError("Not UPPER_SNAKE_CASE!")

    return _Helpers.parse_snake_or_kebab(upper_snake.lower(), "_")

def sentence_to_upper_snake_case(sentence: str) -> str:
    return _Helpers.generate_snake_or_kebab(sentence, "_").upper()

def truncate(string: str, length: int, truncation_string: str = "...", include_suffix: bool = True, remove_whitespace: bool = True) -> str:
    """Truncate string to specified length from the right.

    Args:
        string (str): The string to truncate
        length (int): The length to truncate to
        truncation_string (str, optional): The string to append at the end of truncation. Defaults to "...".
        include_suffix (bool, optional): Include the length of `truncation_string` in the total length of the string. Defaults to True.
        remove_whitespace (bool, optional): Remove all whitespace inbetween the truncated string and truncation_string. Defaults to True.

    Returns:
        str: The truncated string
    """
    target_length = length
    if include_suffix:
        target_length -= len(truncation_string)

    if len(string) <= target_length:
        return string

    txt = string[:target_length]
    if remove_whitespace:
        txt = txt.rstrip()
    return txt + truncation_string

def ltruncate(string: str, length: int, truncation_string: str = "...", include_prefix: bool = True, remove_whitespace: bool = True):
    """Truncate string to specified length from the left.

    Args:
        string (str): The string to truncate
        length (int): The length to truncate to
        truncation_string (str, optional): The string to append at the end of truncation. Defaults to "...".
        include_prefix (bool, optional): Include the length of `truncation_string` in the total length of the string. Defaults to True.
        remove_whitespace (bool, optional): Remove all whitespace inbetween the truncated string and truncation_string. Defaults to True.

    Returns:
        str: The truncated string
    """
    target_length = length
    if include_prefix:
        target_length -= len(truncation_string)

    string_length: int = len(string)
    if string_length <= target_length:
        return string

    start_index = string_length - target_length
    txt = string[start_index:]
    if remove_whitespace:
        txt = txt.lstrip()
    return truncation_string + txt



# Enum imports
from nalpy.console_utils.enums import ConsoleColor as ConsoleColor
from nalpy.console_utils.enums import ConsoleStyle as ConsoleStyle

# Module imports
from nalpy.console_utils import progressbar as progressbar
from nalpy.console_utils import spinner as spinner

# Styling imports
from nalpy.console_utils.styling import set_foreground_color as set_foreground_color
from nalpy.console_utils.styling import set_background_color as set_background_color
from nalpy.console_utils.styling import set_style as set_style
from nalpy.console_utils.styling import reset_attributes as reset_attributes

# Command imports
from nalpy.console_utils.commands import bell as bell
from nalpy.console_utils.commands import backspace as backspace
from nalpy.console_utils.commands import tab as tab
from nalpy.console_utils.commands import line_feed as line_feed
from nalpy.console_utils.commands import form_feed as form_feed
from nalpy.console_utils.commands import carriage_return as carriage_return

# Cursor imports
from nalpy.console_utils.cursor import cursor_up as cursor_up
from nalpy.console_utils.cursor import cursor_down as cursor_down
from nalpy.console_utils.cursor import cursor_forward as cursor_forward
from nalpy.console_utils.cursor import cursor_back as cursor_back
from nalpy.console_utils.cursor import cursor_line_down as cursor_line_down
from nalpy.console_utils.cursor import cursor_line_up as cursor_line_up
from nalpy.console_utils.cursor import cursor_hide as cursor_hide
from nalpy.console_utils.cursor import cursor_show as cursor_show
from nalpy.console_utils.cursor import clear as clear
from nalpy.console_utils.cursor import lclear as lclear
from nalpy.console_utils.cursor import rclear as rclear
from nalpy.console_utils.cursor import erase as erase
from nalpy.console_utils.cursor import lerase as lerase
from nalpy.console_utils.cursor import rerase as rerase
from nalpy.console_utils.cursor import scroll_up as scroll_up
from nalpy.console_utils.cursor import scroll_down as scroll_down
from nalpy.console_utils.cursor import cursor_set_pos as cursor_set_pos
from nalpy.console_utils.cursor import cursor_get_pos as cursor_get_pos



from timeit import timeit
from nalpy import math

STATEMENT = "x % y"

vectorsetup = "x=math.Vector2(69.420, 420.69); y=math.Vector2(-6.9, 6.9)"
floatsetup = "x=69.420; y=-6.9"
glbls = { "math": math }
n = 25_000_000
vec = timeit(STATEMENT, setup=vectorsetup, number=n, globals=glbls)
print(vec)
flt = timeit(STATEMENT, setup=floatsetup, number=n, globals=glbls)
print(flt)
print(f"Float is {vec / flt:.3f} times faster")



import nalpy.console_utils._helpers as _helper

# NOTE: Remember to add to console_utils public imports

def bell():
    """Makes an audible noise."""
    _helper.set_control("\u0007")

def backspace():
    """Moves the cursor one left to enable overwriting of following characters."""
    _helper.set_control("\u0008")

def tab():
    _helper.set_control("\u0009")

def line_feed():
    """Moves to next line, scrolls the display up if at bottom of the screen."""
    _helper.set_control("\u000A")

def form_feed():
    """Move a printer to top of next page."""
    _helper.set_control("\u000C")

def carriage_return():
    """Moves the cursor to column zero."""
    _helper.set_control("\u000D")



import re
from nalpy.humanizer.numbers.enums import NumberFormat
from nalpy.humanizer.numbers import number_to_words_converter, number_to_roman_converter

def _format(num: int, format_: NumberFormat) -> str:
    match format_:
        case NumberFormat.NONE:
            return ""
        case NumberFormat.NUMERALS:
            return str(num)
        case NumberFormat.SPACED_NUMERALS:
            return to_spaced_numerals(num)
        case NumberFormat.WORDS:
            return to_words(num)
        case _:
            raise NotImplementedError(f"No formatter implemented for format '{format_.name}'")

def ordinalize(num: int, format_: NumberFormat = NumberFormat.NUMERALS) -> str:
    if num < 1:
        raise ValueError("Number cannot be less than one!")

    if format_ == NumberFormat.WORDS:
        return to_ordinal_words(num)

    out: str = _format(num, format_)
    last_digit: int = num % 10

    if last_digit == 1:
        return out + "st"
    if last_digit == 2:
        return out + "nd"
    if last_digit == 3:
        return out + "rd"
    return out + "th"

def to_spaced_numerals(num: int):
    reversed_str: str = "".join(reversed(str(num)))
    joined = ' '.join(re.findall(r'.{1,3}', reversed_str))
    return "".join(reversed(joined))

def to_words(num: int, add_and: bool = True) -> str:
    return number_to_words_converter.convert(num, False, add_and)

def to_ordinal_words(num: int) -> str:
    return number_to_words_converter.convert(num, True)

def to_roman(num: int) -> str:
    return number_to_roman_converter.to_roman(num)

def to_percent(num: int | float, format_: NumberFormat = NumberFormat.NUMERALS) -> str:
    if isinstance(num, float):
        num = round(num * 100)

    if format_ == NumberFormat.SPACED_NUMERALS:
        return f"{_format(num, format_)} %"
    if format_ == NumberFormat.WORDS:
        return f"{to_words(num)} percent"
    return f"{_format(num, format_)}%"

def tupleize(count: int) -> str:
    match count:
        case 1:
            return "single"
        case 2:
            return "double"
        case 3:
            return "triple"
        case 4:
            return "quadruple"
        case 5:
            return "quintuple"
        case 6:
            return "sextuple"
        case 7:
            return "septuple"
        case 8:
            return "octuple"
        case 9:
            return "nonuple"
        case 10:
            return "decuple"
        case 100:
            return "centuple"
        case 1000:
            return "milluple"
        case _:
            return f"{count}-tuple"



from typing import Self, final

from .. import hypot, ceil, floor, Vector2Int, Vector2, MVector2

@final
class MVector2Int:
    """A mutable two-dimensional vector with integer precision."""
    __slots__ = "x", "y"

    def __init__(self, x: int, y: int) -> None:
        self.x: int = x
        self.y: int = y

    @classmethod
    def from_immutable(cls, immutable: Vector2Int) -> Self:
        return cls(immutable.x, immutable.y)

    #region Instantiation Helpers
    @classmethod
    def zero(cls) -> Self:
        """Shorthand for ``math.Vector2Int(0, 0)``"""
        return cls(0, 0)

    @classmethod
    def one(cls) -> Self:
        """Shorthand for ``math.Vector2Int(1, 1)``"""
        return cls(1, 1)


    @classmethod
    def up(cls) -> Self:
        """A unit vector pointing up (vector j). Shorthand for ``math.Vector2Int(0, 1)``"""
        return cls(0, 1)

    @classmethod
    def down(cls) -> Self:
        """A unit vector pointing down. Shorthand for ``math.Vector2Int(0, -1)``"""
        return cls(0, -1)

    @classmethod
    def left(cls) -> Self:
        """A unit vector pointing left. Shorthand for ``math.Vector2Int(-1, 0)``"""
        return cls(-1, 0)

    @classmethod
    def right(cls) -> Self:
        """A unit vector pointing right (vector i). Shorthand for ``math.Vector2Int(1, 0)``"""
        return cls(1, 0)
    #endregion

    #region Operators
    def __getitem__(self, i: int) -> int:
        if i == 0:
            return self.x
        if i == 1:
            return self.y
        raise IndexError(i)

    # Provides __str__ also
    def __repr__(self) -> str:
        """Unambiguous string representation of the vector."""
        return f"MVector2Int({self.x}, {self.y})"

    def __add__(self, other: Self | Vector2Int) -> Vector2Int:
        """Add (self + other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__add__(other)

    def __iadd__(self, other: Self | Vector2Int) -> Self:
        """Inline Add (self += other)"""
        if not isinstance(other, MVector2Int | Vector2Int):
            return NotImplemented
        self.x += other.x
        self.y += other.y
        return self

    def __sub__(self, other: Self | Vector2Int) -> Vector2Int:
        """Subtract (self - other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__sub__(other)

    def __isub__(self, other: Self | Vector2Int) -> Self:
        """Inline Subtract (self -= other)"""
        if not isinstance(other, MVector2Int | Vector2Int):
            return NotImplemented
        self.x -= other.x
        self.y -= other.y
        return self

    def __mul__(self, other: Self | Vector2Int | int) -> Vector2Int:
        """Multiply (self * other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__mul__(other)

    def __rmul__(self, other: Self | Vector2Int | int) -> Vector2Int:
        """Reverse multiply (other * self)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__rmul__(other)

    def __imul__(self, other: Self | Vector2Int | int) -> Self:
        """Inline Multiply (self *= other)"""
        x: int
        y: int
        if isinstance(other, MVector2Int | Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        self.x *= x
        self.y *= y
        return self

    def __truediv__(self, other: Self | Vector2Int | int | float) -> Vector2:
        """Divide (self / other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__truediv__(other)

    def __floordiv__(self, other: Self | Vector2Int | int) -> Vector2Int:
        """Floor Divide (self // other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__floordiv__(other)

    def __ifloordiv__(self, other: Self | Vector2Int | int) -> Self:
        """Inline Floor Divide (self //= other)"""
        x: int
        y: int
        if isinstance(other, MVector2Int | Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        self.x //= x
        self.y //= y
        return self


    def __mod__(self, other: Self | Vector2Int | int) -> Vector2Int:
        """Modulo (self % other)"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__mod__(other)

    def __divmod__(self, other: Self | Vector2Int | int) -> tuple[Vector2Int, Vector2Int]:
        """Floor division and modulo (divmod(self, other))"""
        if isinstance(other, MVector2Int):
            other = other.to_immutable()
        return self.to_immutable().__divmod__(other)


    def __eq__(self, other: Self) -> bool:
        if isinstance(other, MVector2Int):
            return self.x == other.x and self.y == other.y
        raise TypeError(f"Cannot compare equality between {type(self)} and {type(other)}")

    # Hash not implemented because object is mutable.

    #endregion

    #region Instance Properties
    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        return hypot(self.x, self.y)
    #endregion

    #region Operations
    def copy(self) -> Self:
        """Make a new vector where the components' values match this vectors values."""
        return MVector2Int(self.x, self.y)
    #endregion

    #region Constructors
    @classmethod
    def ceil(cls, v: MVector2) -> Self:
        return cls(ceil(v.x), ceil(v.y))

    @classmethod
    def floor(cls, v: MVector2) -> Self:
        return cls(floor(v.x), floor(v.y))

    @classmethod
    def round(cls, v: MVector2) -> Self:
        return cls(round(v.x), round(v.y))

    @classmethod
    def min(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the smallest components of two vectors."""
        return cls(min(a.x, b.x), min(a.y, b.y))

    @classmethod
    def max(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the largest components of two vectors."""
        return cls(max(a.x, b.x), max(a.y, b.y))
    #endregion

    #region Conversions
    def to_mvector2(self) -> MVector2:
        return MVector2(float(self.x), float(self.y))

    def to_immutable(self) -> Vector2Int:
        return Vector2Int(self.x, self.y)

    def to_int_dict(self) -> dict[str, int]:
        return {"x": self.x, "y": self.y}
    #endregion



from typing import NamedTuple, Self, final, Final

from .. import hypot, clamp01, sqrt, clamp, acos, degrees, sign

@final
class Vector2(NamedTuple):
    """An immutable two-dimensional vector"""
    x: float
    y: float

    #region Class Properties
    @classmethod
    @property
    def zero(cls) -> Self:
        """Shorthand for ``math.Vector2(0.0, 0.0)``"""
        return _ZERO #  Returning single instance, because Vector2 is immutable

    @classmethod
    @property
    def one(cls) -> Self:
        """Shorthand for ``math.Vector2(1.0, 1.0)``"""
        return _ONE #  Returning single instance, because Vector2 is immutable


    @classmethod
    @property
    def up(cls) -> Self:
        """A unit vector pointing up (vector j). Shorthand for ``math.Vector2(0.0, 1.0)``"""
        return _UP #  Returning single instance, because Vector2 is immutable

    @classmethod
    @property
    def down(cls) -> Self:
        """A unit vector pointing down. Shorthand for ``math.Vector2(0.0, -1.0)``"""
        return _DOWN #  Returning single instance, because Vector2 is immutable

    @classmethod
    @property
    def left(cls) -> Self:
        """A unit vector pointing left. Shorthand for ``math.Vector2(-1.0, 0.0)``"""
        return _LEFT #  Returning single instance, because Vector2 is immutable

    @classmethod
    @property
    def right(cls) -> Self:
        """A unit vector pointing right (vector i). Shorthand for ``math.Vector2(1.0, 0.0)``"""
        return _RIGHT #  Returning single instance, because Vector2 is immutable
    #endregion

    #region Operators

    # Provides __str__ also
    def __repr__(self) -> str:
        """Unambiguous string representation of the vector."""
        return f"Vector2({self.x}, {self.y})"

    def __add__(self, other: Self) -> Self:
        """Add (self + other)"""
        if not isinstance(other, Vector2):
            return NotImplemented
        return Vector2(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Self) -> Self:
        """Subtract (self - other)"""
        if not isinstance(other, Vector2):
            return NotImplemented
        return Vector2(self.x - other.x, self.y - other.y)

    def __mul__(self, other: Self | float | int) -> Self:
        """Multiply (self * other)"""
        x: float | int
        y: float | int
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2(self.x * x, self.y * y)

    def __rmul__(self, other: Self | float | int) -> Self:
        """Reverse multiply (other * self)"""
        return self.__mul__(other) # other * self = self * other

    def __truediv__(self, other: Self | float | int) -> Self:
        """Divide (self / other)"""
        x: float | int
        y: float | int
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2(self.x / x, self.y / y)

    def __floordiv__(self, other: Self | float | int) -> Self:
        """Floor Divide (self // other)"""
        x: float | int
        y: float | int
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2(self.x // x, self.y // y)


    def __mod__(self, other: Self | float | int) -> Self:
        """Modulo (self % other)"""
        x: float | int
        y: float | int
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2(self.x % x, self.y % y)

    def __divmod__(self, other: Self | float | int) -> tuple[Self, Self]:
        """Floor division and modulo (divmod(self, other))"""
        x: float | int
        y: float | int
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        x_fdiv, x_mod = divmod(self.x, x)
        y_fdiv, y_mod = divmod(self.y, y)
        return (Vector2(x_fdiv, y_fdiv), Vector2(x_mod, y_mod))


    def __neg__(self) -> Self:
        """Negate (-self)"""
        return Vector2(-self.x, -self.y)

    def __abs__(self) -> Self:
        """Absolute value (abs(self))"""
        return Vector2(abs(self.x), abs(self.y))

    # __eq__ and __hash__ are provided by NamedTuple

    #endregion

    #region Instance Properties
    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        return hypot(self.x, self.y)

    @property
    def normalized(self) -> Self:
        """A copy of this vector with a magnitude of 1"""
        mag: float = self.magnitude
        if mag == 0:
            raise ValueError("Vector does not have a direction to normalize to.")
        return Vector2(self.x / mag, self.y / mag)
    #endregion

    #region Mathematic Operations
    @classmethod
    def dot(cls, a: Self, b: Self) -> float:
        """Dot Product of two vectors."""
        return (a.x * b.x) + (a.y * b.y)
    #endregion

    #region Interpolation
    @classmethod
    def lerp(cls, a: Self, b: Self, t: float) -> Self:
        """
        Linearly interpolates between vectors ``a`` and ``b`` by ``t``.

        The parameter ``t`` is clamped to the range [0, 1].
        """
        t = clamp01(t)
        lerp_x = a.x + ((b.x - a.x) * t)
        lerp_y = a.y + ((b.y - a.y) * t)
        return Vector2(lerp_x, lerp_y)

    @classmethod
    def lerp_unclamped(cls, a: Self, b: Self, t: float) -> Self:
        """
        Linearly interpolates between vectors ``a`` and ``b`` by ``t``.

        The parameter ``t`` is not clamped.
        """
        lerp_x = a.x + ((b.x - a.x) * t)
        lerp_y = a.y + ((b.y - a.y) * t)
        return Vector2(lerp_x, lerp_y)

    @classmethod
    def move_towards(cls, current: Self, target: Self, max_distance_delta: float):
        """Moves a point current towards target."""
        to_vector_x: float = target.x - current.x
        to_vector_y: float = target.y - current.y

        sqDist: float = to_vector_x * to_vector_x + to_vector_y * to_vector_y

        if sqDist == 0 or (max_distance_delta >= 0 and sqDist <= max_distance_delta * max_distance_delta):
            return target

        dist: float = sqrt(sqDist)

        move_x = current.x + to_vector_x / dist * max_distance_delta
        move_y = current.y + to_vector_y / dist * max_distance_delta
        return Vector2(move_x, move_y)
    #endregion

    #region Rotation
    @classmethod
    def perpendicular(cls, vector: Self) -> Self:
        """
        Returns a 2D vector with the same magnitude, but perpendicular to the given 2D vector.
        The result is always rotated 90-degrees in a counter-clockwise direction for a 2D coordinate system where the positive Y axis goes up.
        """
        return Vector2(-vector.y, vector.x)

    @classmethod
    def reflect(cls, vector: Self, normal: Self):
        """Reflects a vector off the vector defined by a normal."""
        factor: float = -2 * Vector2.dot(normal, vector)
        return Vector2(factor * normal.x + vector.x, factor * normal.y + vector.y)
    #endregion

    #region Relation
    @classmethod
    def angle(cls, _from: Self, _to: Self) -> float:
        """
        Gets the unsigned angle in degrees between from and to.

        NOTE: The angle returned will always be between 0 and 180 degrees, because the method returns the smallest angle between the vectors.
        """
        denom: float = _from.magnitude * _to.magnitude
        if denom == 0.0:
            return 0.0

        dot: float = Vector2.dot(_from, _to)
        cos_val: float = clamp(dot / denom, -1.0, 1.0)
        return degrees(acos(cos_val))

    @classmethod
    def signed_angle(cls, _from: Self, _to: Self) -> float:
        """
        Gets the signed angle in degrees between from and to. The angle returned is the signed counterclockwise angle between the two vectors.

        NOTE: The angle returned will always be between -180 and 180 degrees, because the method returns the smallest angle between the vectors.
        """
        unsigned_angle: float = Vector2.angle(_from, _to)
        s: float = sign((_from.x * _to.y) - (_from.y * _to.x))
        return unsigned_angle * s

    @classmethod
    def distance(cls, a: Self, b: Self):
        """Returns the distance between a and b."""
        diff = a - b
        return diff.magnitude
    #endregion

    #region Contructors
    @classmethod
    def min(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the smallest components of two vectors."""
        return cls(min(a.x, b.x), min(a.y, b.y))

    @classmethod
    def max(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the largest components of two vectors."""
        return cls(max(a.x, b.x), max(a.y, b.y))
    #endregion

    #region Conversions
    def to_int_tuple(self) -> tuple[int, int]:
        return (round(self.x), round(self.y))

    def to_float_dict(self) -> dict[str, float]:
        return self._asdict()
    #endregion

_ZERO: Final[Vector2] = Vector2(0.0, 0.0)
_ONE: Final[Vector2] = Vector2(1.0, 1.0)

_UP: Final[Vector2] = Vector2(0.0, 1.0)
_DOWN: Final[Vector2] = Vector2(0.0, -1.0)
_LEFT: Final[Vector2] = Vector2(-1.0, 0.0)
_RIGHT: Final[Vector2] = Vector2(1.0, 0.0)



from setuptools import setup, Extension, find_packages

setup(
    packages=find_packages(where=".", include=["nalpy*"]),
    ext_modules=[
        Extension("nalpy.math._c_extensions.functions", ["nalpy/math/_c_extensions/functions.c"]),
        Extension("nalpy.math._c_extensions.vector2", ["nalpy/math/_c_extensions/vector2.c"]),
        Extension("nalpy.math._c_extensions.vector2_int", ["nalpy/math/_c_extensions/vector2_int.c"]),
        Extension("nalpy.math._c_extensions.mvector2", ["nalpy/math/_c_extensions/mvector2.c"]),
        Extension("nalpy.math._c_extensions.mvector2_int", ["nalpy/math/_c_extensions/mvector2_int.c"])
    ],
    package_data={"nalpy": ["py.typed"], "nalpy.math._c_extensions": ["*.pyi"]},
    zip_safe=False
)



from typing import Final, NamedTuple, Self, final

from .. import lerp, inverse_lerp, Vector2


@final
class Rect(NamedTuple):
    """
    A 2D Rectangle defined by X and Y position, width and height.

    The functionality of this rect depends on the coordinate space used.
    When X increases to the right and Y increases upwards, position represents the bottom-left corner of the rectangle.
    When X increases to the right and Y increases downwards, position represents the top-left corner of the rectange.
    """

    x: float
    y: float
    w: float
    h: float

    @classmethod
    @property
    def zero(cls) -> Self:
        """Shorthand for ``math.Rect(0.0, 0.0, 0.0, 0.0)``"""
        return _ZERO #  Returning single instance, because Rect is immutable

    #region Constructors
    @classmethod
    def from_pos(cls, position: Vector2, size: Vector2) -> Self:
        return cls(position.x, position.y, size.x, size.y)

    @classmethod
    def from_center(cls, center: Vector2, size: Vector2) -> Self:
        x: float = center.x - (size.x / 2)
        y: float = center.y - (size.y / 2)
        return cls(x, y, size.x, size.y)

    @classmethod
    def from_min_max(cls, xmin: float, ymin: float, xmax: float, ymax: float) -> Self:
        w: float = xmax - xmin
        h: float = ymax - ymin

        return cls(xmin, ymin, w, h)

    @classmethod
    def from_min_max_vectors(cls, min: Vector2, max: Vector2) -> Self:
        return cls.from_min_max(min.x, min.y, max.x, max.y)
    #endregion

    #region Operators

    # Provides __str__ also
    def __repr__(self) -> str:
        return f"Rect({self.x}, {self.y}, {self.w}, {self.h})"
    #endregion

    #region Properties
    @property
    def center(self) -> Vector2:
        return Vector2(self.xcenter, self.ycenter)
    @property
    def xcenter(self) -> float:
        return self.x + (self.w / 2.0)
    @property
    def ycenter(self) -> float:
        return self.y + (self.h / 2.0)

    @property
    def min(self) -> Vector2:
        """Same as `Rect.position`"""
        return Vector2(self.xmin, self.ymin)
    @property
    def xmin(self) -> float:
        """Same as `Rect.x`"""
        return self.x
    @property
    def ymin(self) -> float:
        """Same as `Rect.y`"""
        return self.y

    @property
    def max(self) -> Vector2:
        return Vector2(self.xmax, self.ymax)
    @property
    def xmax(self) -> float:
        return self.x + self.w
    @property
    def ymax(self) -> float:
        return self.y + self.h

    @property
    def position(self) -> Vector2:
        return Vector2(self.x, self.y)
    @property
    def size(self) -> Vector2:
        return Vector2(self.w, self.h)
    #endregion

    #region Collisions
    def contains(self, point: Vector2) -> bool:
        return (point.x >= self.x) and (point.x < self.xmax) and (point.y >= self.y) and (point.y < self.ymax)

    def overlaps(self, other: Self) -> bool:
        return (
            other.xmax > self.x and
            other.x    < self.xmax and
            other.ymax > self.y and
            other.y    < self.ymax
        )
    #endregion

    #region Normalization
    @classmethod
    def normalized_to_point(cls, rect: Self, normalized_rect_coordinates: Vector2) -> Vector2:
        return Vector2(
            lerp(rect.x, rect.xmax, normalized_rect_coordinates.x),
            lerp(rect.y, rect.ymax, normalized_rect_coordinates.y)
        )

    @classmethod
    def point_to_normalized(cls, rect: Self, point: Vector2) -> Vector2:
        return Vector2(
            inverse_lerp(rect.x, rect.xmax, point.x),
            inverse_lerp(rect.y, rect.ymax, point.y)
        )
    #endregion

    #region Conversions
    def to_int_tuple(self) -> tuple[int, int, int, int]:
        return (round(self.x), round(self.y), round(self.w), round(self.h))

    def to_float_dict(self) -> dict[str, float]:
        return self._asdict()
    #endregion

_ZERO: Final[Rect] = Rect(0.0, 0.0, 0.0, 0.0)



from enum import Enum
from nalpy import math
from typing import Iterable, NamedTuple, Self, TypeVar


class _Scale(NamedTuple):
    name: str
    symbol: str
    value_base: int
    value_exp: int

    @property
    def value(self) -> int:
        return pow(self.value_base, self.value_exp)

    @classmethod
    def create_empty_scale(cls, base: int) -> Self:
        return cls(name="", symbol="", value_base=base, value_exp=0)


class MetricPrefix(Enum):
    QUETTA = _Scale("quetta", "Q",  10,  30)
    RONNA  = _Scale("ronna",  "R",  10,  27)
    YOTTA  = _Scale("yotta",  "Y",  10,  24)
    ZETTA  = _Scale("zetta",  "Z",  10,  21)
    EXA    = _Scale("exa",    "E",  10,  18)
    PETA   = _Scale("peta",   "P",  10,  15)
    TERA   = _Scale("tera",   "T",  10,  12)
    GIGA   = _Scale("giga",   "G",  10,  9)
    MEGA   = _Scale("mega",   "M",  10,  6)
    KILO   = _Scale("kilo",   "k",  10,  3)
    HECTO  = _Scale("hecto",  "h",  10,  2)
    DECA   = _Scale("deca",   "da", 10,  1)

    DECI   = _Scale("deci",   "d",  10, -1)
    CENTI  = _Scale("centi",  "c",  10, -2)
    MILLI  = _Scale("milli",  "m",  10, -3)
    MICRO  = _Scale("micro",  "μ",  10, -6)
    NANO   = _Scale("nano",   "n",  10, -9)
    PICO   = _Scale("pico",   "p",  10, -12)
    FEMTO  = _Scale("femto",  "f",  10, -15)
    ATTO   = _Scale("atto",   "a",  10, -18)
    ZEPTO  = _Scale("zepto",  "z",  10, -21)
    YOCTO  = _Scale("yocto",  "y",  10, -24)
    RONTO  = _Scale("ronto",  "r",  10, -27)
    QUECTO = _Scale("quecto", "q",  10, -30)

    @property
    def value(self) -> _Scale:
        val = super().value
        assert isinstance(val, _Scale)
        return val

    @classmethod
    def get_optimal_for_value(cls, base_value: float) -> Self | None:
        return _pick_optimal_prefix(base_value, cls)

class BinaryPrefix(Enum):
    KIBI = _Scale("kibi", "Ki", 2, 10)
    MEBI = _Scale("mebi", "Mi", 2, 20)
    GIBI = _Scale("gibi", "Gi", 2, 30)
    TEBI = _Scale("tebi", "Ti", 2, 40)
    PEBI = _Scale("pebi", "Pi", 2, 50)
    EXBI = _Scale("exbi", "Ei", 2, 60)
    ZEBI = _Scale("zebi", "Zi", 2, 70)
    YOBI = _Scale("yobi", "Yi", 2, 80)

    @property
    def value(self) -> _Scale:
        val = super().value
        assert isinstance(val, _Scale)
        return val

    @classmethod
    def get_optimal_for_value(cls, base_value: float) -> Self | None:
        return _pick_optimal_prefix(base_value, cls)

_PrefixT = TypeVar("_PrefixT", MetricPrefix, BinaryPrefix)

def convert_metric(input_: float, input_scale: MetricPrefix | BinaryPrefix | None, output_scale: MetricPrefix | BinaryPrefix | None) -> float:
    i_scale: _Scale
    if input_scale is not None:
        i_scale = input_scale.value
    elif output_scale is not None:
        i_scale = _Scale.create_empty_scale(output_scale.value.value_base)
    else:
        i_scale = _Scale.create_empty_scale(1)

    o_scale: _Scale = output_scale.value if output_scale is not None else _Scale.create_empty_scale(i_scale.value_base)

    scale_multiplier: float
    if i_scale.value_base == o_scale.value_base:
        scale_multiplier = math.pow(i_scale.value_base, i_scale.value_exp - o_scale.value_exp)
    elif i_scale.value_exp == o_scale.value_exp:
        scale_multiplier = math.pow(i_scale.value_base / o_scale.value_base, i_scale.value_exp)
    else:
        scale_multiplier = i_scale.value / o_scale.value

    return input_ * scale_multiplier

def to_metric(base_value: float, unit_prefix: MetricPrefix | BinaryPrefix | None, unit_name: str | None = None, with_space: bool = True, decimals: int = 0) -> str:
    scale_value: int
    scale_name: str
    if unit_prefix is not None:
        scale_value = unit_prefix.value.value
        scale_name = unit_prefix.value.name
    else:
        scale_value = 1
        scale_name = ""
    return _to_metric(base_value, scale_value, scale_name, unit_name, with_space, decimals, True)

def to_short_metric(base_value: float, unit_prefix: MetricPrefix | BinaryPrefix | None, unit_symbol: str | None = None, with_space: bool = True, decimals: int = 0) -> str:
    scale_value: int
    scale_symbol: str
    if unit_prefix is not None:
        scale_value = unit_prefix.value.value
        scale_symbol = unit_prefix.value.symbol
    else:
        scale_value = 1
        scale_symbol = ""
    return _to_metric(base_value, scale_value, scale_symbol, unit_symbol, with_space, decimals, False)

def _to_metric(base_value: float, scale_value: int, scale_name: str, unit: str | None, with_space: bool, decimals: int, use_plural: bool) -> str:
    space: str = " " if with_space else ""
    value = math.round_to_digits(base_value / scale_value, digits=decimals)

    unit_str: str
    if unit is None:
        unit_str = ""
    elif (decimals == 0 and int(value) == 1) or not use_plural:
        unit_str = unit
    else:
        unit_str = unit + "s"

    return f"{value:.{decimals}f}{space}{scale_name}{unit_str}"

def _pick_optimal_prefix(base_value: float, prefixes: Iterable[_PrefixT]) -> _PrefixT | None:
    def get_prefix_value(prefix: _PrefixT | None) -> int:
        if prefix is None:
            return 1
        return prefix.value.value

    prefix: _PrefixT | None = None

    for prefix in sorted((*prefixes, None), key=lambda s: get_prefix_value(s), reverse=True):
        scale_value = get_prefix_value(prefix)
        if (base_value / scale_value) >= 1.0:
            return prefix

    return prefix



#cython: language_level=3

from libc.math cimport llround, fabs

def round(double x, /):
    return llround(x)

def remap(double value, double from1, double to1, double from2, double to2):
    return (value - from1) / (to1 - from1) * (to2 - from2) + from2

def remap01(double value, double from1, double to1):
    return (value - from1) / (to1 - from1)

def clamp(value, _min, _max): # types not verified due to performance reasons
    if value < _min:
        return _min
    if value > _max:
        return _max
    return value

cpdef clamp01(double value) noexcept:
    if value < 0.0:
        return 0.0
    if value > 1.0:
        return 1.0
    return value

cpdef delta_angle(double current, double target) noexcept:
    cdef double delta = (target - current) % 360.0
    if delta > 180.0:
        delta -= 360.0
    return delta

def sign(double x, /):
    return doublesign(x)


def lerp(double a, double b, double t):
    return a + (b - a) * clamp01(t)

def lerp_unclamped(double a, double b, double t):
    return a + (b - a) * t

def lerp_angle(double a, double b, double t):
    return a + delta_angle(a, b) * clamp01(t)

def inverse_lerp(double a, double b, double value):
    if a != b:
        return clamp01((value - a) / (b - a))
    else:
        return 0.0

def smooth_step(double a, double b, double t):
    t = clamp01(t)
    t = -2.0 * t * t * t + 3.0 * t * t
    return b * t + a * (1 - t)

cpdef move_towards(double current, double target, double max_delta) noexcept:
    if fabs(target - current) <= max_delta:
        return target
    return current + doublesign(target - current) * max_delta

def move_towards_angle(double current, double target, double max_delta):
    deltaAngle = delta_angle(current, target)
    if -max_delta < deltaAngle and deltaAngle < max_delta:
        return target
    target = current + deltaAngle
    return move_towards(current, target, max_delta)

def ping_pong(double t, double length):
    t = t % (length * 2.0)
    return length - fabs(t - length)



import typing as _typing

_T = _typing.TypeVar("_T")

def humanize_iterable(iterable: _typing.Iterable[_T], separator: str = "and", formatter: _typing.Callable[[_T], str] | None = None) -> str:
    if formatter is None:
        formatter = str

    return _humanize_display_strings((formatter(o) for o in iterable), separator)

def _humanize_display_strings(strings: _typing.Iterable[str], separator: str) -> str:
    items: tuple[str, ...] = tuple(s for s in strings if len(s) > 0 and not s.isspace())

    count: int = len(items)

    if count == 0:
        return ""

    if count == 1:
        return items[0]

    before_last: tuple[str, ...] = items[:-1]
    last: str = items[-1]

    before_last_joined: str = ", ".join(before_last)

    return _get_conjunction_format_string(count).format(before_last=before_last_joined, sep=separator, last=last)

def _get_conjunction_format_string(item_count: int) -> str:
    if item_count > 2:
        return "{before_last}, {sep} {last}"

    return "{before_last} {sep} {last}"




from nalpy.humanizer import bytes as bytes
from nalpy.humanizer import inflections as inflections
from nalpy.humanizer import iterable as iterable
from nalpy.humanizer import numbers as numbers
from nalpy.humanizer import string as string
from nalpy.humanizer import time as time
